/* 
Antennenkopplersteuerung 
Autor:    Roland Kaiser, OE2ROL
          Michael Lueftenegger, OE2LWP
Hardware: Arduino Mega 2560

Die Autoren übernehmen keine Haftung etwaiger Folgeschäden an Transceivern oder dgl.
   
Diese Software ist für den eigenen privaten Gebrauch frei verfügbar, jedoch nicht für kommerzielle Zwecke ohne Genehmigung der Autoren.
Veränderungen und Verbesserungen der Software bitte an Roland Kaiser OE2ROL (roland.kaiser@sbg.at) senden.

Anschlussbelegung der Ports:

CI-V port:                     (Serial 1)
RX                             Arduino Pin D19
TX                             Arduino Pin D18

Nextion Serial Port:           (Serial 3)
RX                             Arduino Pin D15
TX                             Arduino Pin D14

SD Card Reader SPI Port:
CS ->                          Arduino Pin D53
MOSI ->                        Arduino Pin D51
MISO ->                        Arduino Pin D50
SCK ->                         Arduino Pin D52

One Wire Port->                Arduino Pin D23

*/

#include <SPI.h>
#include <SD.h>

#include "LookupTable.h"
#include "Display.h"
#include "Transceiver.h"
#include "Koppler.h"
#include "Config.h"

unsigned long stepFrequency;
unsigned long step_lc;
bool resetRequired = false;
LookupTable lookuptable = LookupTable(Config::FILENAME_FREQ);
Display display = Display();
Transceiver transceiver = Transceiver(Config::CONTROLLER_ADDR, Config::CIV_ADDR);
Koppler koppler = Koppler();

bool forceUpdate = true;

void reset() {
  display.showPage(Config::NAME_PAGE_SPLASH);
  delay(100);
  //soft reset
  asm volatile("  jmp 0");
}

void invalidate() {
  forceUpdate = true;
}

String format(String value, unsigned char length) {
  String result = "";
  for (int i = 0; i < length - value.length(); i++)
    result += " ";
  return value + result;
}

void showHelp() {
  unsigned char size = 20;
  Serial.println("Kopplersteuerung von OE2ROL.");
  Serial.println("");
  Serial.println(format("p0", size)                 + " : " + "shows page 0");
  Serial.println(format("p1", size)                 + " : " + "shows page 1");
  Serial.println(format("p2", size)                 + " : " + "shows page 2");
  Serial.println(format("p3", size)                 + " : " + "shows page 3");
  Serial.println(format("p4", size)                 + " : " + "shows page 4");
  Serial.println(format("p5", size)                 + " : " + "shows page 5");
  Serial.println(format("ps", size)                 + " : " + "shows Splash screen");
  Serial.println(format("pw", size)                 + " : " + "shows Wait screen");
  Serial.println(format("pc", size)                 + " : " + "shows Config screen");
  Serial.println(format("kb", size)                 + " : " + "shows keyboard");
  Serial.println(format("info", size)               + " : " + "shows the current state.");
  Serial.println(format("bit", size)                + " : " + "shows the bit pattern of all chips as nibbles.");
  Serial.println(format("sf", size)                 + " : " + "sets the frequency in Hz");
  Serial.println(format("dump [freq]", size)        + " : " + "shows the content of the lookup table. Or the dump of the given frequency.");
  Serial.println(format("append freq c l io", size) + " : " + "append a line in the lookuptable. ! One space !");
  Serial.println(format("update", size)             + " : " + "updates the system state.");
  Serial.println(format("config", size)             + " : " + "shows config values.");
  Serial.println(format("config key value", size)   + " : " + "sets the config value for 'key' to 'value'.");
  Serial.println(format("config load", size)        + " : " + "loads config values.");
  Serial.println(format("config save", size)        + " : " + "saves config values.");
  Serial.println(format("config dump", size)        + " : " + "shows the config file.");
  Serial.println(format("help", size)               + " : " + "shows this help page.");
  Serial.println(format("reset", size)              + " : " + "executes a reset.");
  /**/
}

void binPrint(int var) {
  for (int i = 0; i < 8; i++) {
    Serial.print(((var >> 7 - i) & 1) == 1 ? "1" : "0");
  }
  Serial.println();
}

void binPrint1(int b1, int b2, int b3) {
  Serial.println("8      1|8      1|8      1");
  for (int i = 0; i < 8; i++) {
    Serial.print(((b1 >> 7 - i) & 1) == 1 ? "1" : "0");
  }
  Serial.print(" ");

  for (int i = 0; i < 8; i++) {
    Serial.print(((b2 >> 7 - i) & 1) == 1 ? "1" : "0");
  }
  Serial.print(" ");

  for (int i = 0; i < 8; i++) {
    Serial.print(((b3 >> 7 - i) & 1) == 1 ? "1" : "0");
  }
  Serial.println("");
}

void showInfo() {
  unsigned char size = 25;
  Serial.println("Info");
  Serial.println("--------------------------------------");
  Serial.println(format("VERSION", size)                 + " : " + Config::VERSION);
  Serial.println(format("DATE", size)                    + " : " + Config::DATE);
  Serial.println(format("Step L,C", size)                + " : " + String(step_lc));
  Serial.println(format("Step frequency", size)          + " : " + String(stepFrequency));
  Serial.println(format("Transceiver connected", size)   + " : " + String(transceiver.isConnected()));
  Serial.println(format("Request interval", size)        + " : " + String(transceiver.getRequestInterval()));
  Serial.println(format("Tranceiver frequency", size)    + " : " + String(transceiver.getFrequency()));
  Serial.println(format("Tranceiver frequency(r)", size) + " : " + String(transceiver.getFrequencyRounded()));
  Serial.println(format("C", size)                       + " : " + String(koppler.getC()) + "(" + String(koppler.getC() * Config::STEP_C_MIN_VALUE) + ")");
  Serial.println(format("L", size)                       + " : " + String(koppler.getL()) + "(" + String(koppler.getL() * Config::STEP_L_MIN_VALUE) + ")");
  Serial.println(format("IO", size)                      + " : " + String(koppler.getIO()));
  Serial.print(format("IC1", size)                       + " : " + String(koppler.getIC1()) + " : ");
  binPrint(koppler.getIC1());
  Serial.print(format("IC2", size)                       + " : " + String(koppler.getIC2()) + " : ");
  binPrint(koppler.getIC2());
  Serial.print(format("IC3", size)                       + " : " + String(koppler.getIC3()) + " : ");
  binPrint(koppler.getIC3());
}

void showBit() {
  unsigned char size = 25;

  binPrint1(koppler.getIC1(), koppler.getIC2(), koppler.getIC3());
}

void showConfig() {
  unsigned char size = 30;
  Serial.println("CONFIG VALUES");
  Serial.println("-------------");
  Serial.println(format("initial_frequency*", size)         + " : " + String(Config::INITIAL_FREQUENCY));
  Serial.println(format("request_interval*", size)          + " : " + String(transceiver.getRequestInterval()));
  Serial.println(format("brightness*", size)                + " : " + String(Config::BRIGHTNESS));
  Serial.println(format("restart_delay*", size)             + " : " + String(Config::RESTART_DELAY));

  Serial.print(format("controller_addr*(HEX)", size)        + " : ");
  Serial.println(Config::CONTROLLER_ADDR, HEX);

  Serial.print(format("civ_addr*(HEX)", size)               + " : ");
  Serial.println(Config::CIV_ADDR, HEX);

  Serial.println(format("invert_koppler_output_c*", size)   + " : " + String(Config::INVERT_KOPPLER_OUTPUT_C));
  Serial.println(format("invert_koppler_output_l*", size)   + " : " + String(Config::INVERT_KOPPLER_OUTPUT_L));
  Serial.println(format("invert_koppler_output_io*", size)  + " : " + String(Config::INVERT_KOPPLER_OUTPUT_IO));
  Serial.println(format("scroll_width*", size)              + " : " + String(Config::SCROLL_WIDTH));
  Serial.println(format("list_size*", size)                 + " : " + String(Config::LIST_SIZE));
  Serial.println(format("step_bit_width_l*", size)          + " : " + String(Config::STEP_BIT_WIDTH_L));
  Serial.println(format("step_bit_width_c*", size)          + " : " + String(Config::STEP_BIT_WIDTH_C));
  Serial.println(format("step_c_min_value*", size)          + " : " + String(Config::STEP_C_MIN_VALUE, 3));
  Serial.println(format("step_l_min_value*", size)          + " : " + String(Config::STEP_L_MIN_VALUE, 3));
  Serial.println(format("pin_onewire*", size)               + " : " + String(Config::PIN_ONEWIRE));

  Serial.print(format("ic1*(HEX)", size) + " : ");
  for (unsigned char i = 0; i < 8; i++) {
    Serial.print(Config::IC1[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  Serial.print(format("ic2*(HEX)", size) + " : ");
  for (unsigned char i = 0; i < 8; i++) {
    Serial.print(Config::IC2[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  Serial.print(format("ic3*(HEX)", size) + " : ");
  for (unsigned char i = 0; i < 8; i++) {
    Serial.print(Config::IC3[i], HEX);
    Serial.print(" ");
  }
  Serial.println("");

  Serial.println("");
  Serial.println(format("CHIP_SELECT_PIN", size)              + " : " + String(Config::CHIP_SELECT_PIN));
  Serial.println(format("SERIAL_SPEED_SERIAL1", size)         + " : " + String(Config::SERIAL_SPEED_SERIAL1));
  Serial.println(format("SERIAL_SPEED_SERIAL2", size)         + " : " + String(Config::SERIAL_SPEED_SERIAL2));
  Serial.println(format("SERIAL_SPEED_SERIAL3", size)         + " : " + String(Config::SERIAL_SPEED_SERIAL3));
  Serial.println(format("STEP_LC_INITIAL_VALUE", size)        + " : " + String(Config::STEP_LC_INITIAL_VALUE));
  Serial.println(format("STEP_FREQUENCY_INITIAL_VALUE", size) + " : " + String(Config::STEP_FREQUENCY_INITIAL_VALUE));
  Serial.println(format("FILENAME_FREQ", size)                + " : " + Config::FILENAME_FREQ);
  Serial.println(format("FILENAME_CONFIG", size)              + " : " + Config::FILENAME_CONFIG);

  Serial.println("*) value can be set with config command.");
}

String getPart(String cmd, unsigned char part) {
  String delimiter = " ";
  int posStart = 0;
  int posEnd = cmd.length();

  unsigned char partCount = 0;
  for (int i = 0; i <= cmd.length(); i++) {
    if (cmd.substring(i, i + 1) != delimiter) {
      posStart = i;
      if (partCount == part) {
        break;
      }
    }
    if (cmd.substring(i, i + 1) == delimiter)
      partCount++;
  }

  for (int i = posStart + 1; i < cmd.length(); i++) {
    if (cmd.substring(i, i + 1) == delimiter) {
      posEnd = i;
      break;
    }
  }
  if (posStart >= posEnd)
    return "";

  return cmd.substring(posStart, posEnd);
}

void console() {
  String cmd;
  if (Serial.available()) {

    cmd = Serial.readString();
    cmd.trim();
    cmd.toLowerCase();

    Serial.println("echo:" + cmd);

    if (getPart(cmd, 0) == "p0") {
      display.showPage(Config::NAME_PAGE_AUTO_LEFT);
    } else if (getPart(cmd, 0) == "p1") {
      display.showPage(Config::NAME_PAGE_AUTO_RIGHT);
    } else if (getPart(cmd, 0) == "p2") {
      display.showPage(Config::NAME_PAGE_FREQ_LEFT);
    } else if (getPart(cmd, 0) == "p3") {
      display.showPage(Config::NAME_PAGE_FREQ_RIGHT);
    } else if (getPart(cmd, 0) == "p4") {
      display.showPage(Config::NAME_PAGE_MAN_C_LEFT);
    } else if (getPart(cmd, 0) == "p5") {
      display.showPage(Config::NAME_PAGE_MAN_C_RIGHT);
    } else if (getPart(cmd, 0) == "ps") {
      display.showPage(Config::NAME_PAGE_SPLASH);
    } else if (getPart(cmd, 0) == "pw") {
      display.showPage(Config::NAME_PAGE_WAIT);
    } else if (getPart(cmd, 0) == "pc") {
      display.showConfig();
    } else if (getPart(cmd, 0) == "kb") {
      display.showPage(Config::NAME_PAGE_KEYBOARD);
    } else if (getPart(cmd, 0) == "reset") {
      reset();
    } else if (getPart(cmd, 0) == "info") {
      showInfo();
    } else if (getPart(cmd, 0) == "bit") {
      showBit();
    } else if (getPart(cmd, 0) == "sf") {
      String value = getPart(cmd, 1);
      value.trim();
      transceiver.setFrequency(value.toInt());
    } else if (cmd == "config") {
      showConfig();
    } else if (getPart(cmd, 0) == "config") {

      String key = getPart(cmd, 1);

      if (key == "dump") {
        Config::dump();
        return;
      } else if (key == "save") {
        Config::save();
        return;
      } else if (key == "load") {
        Config::load();
        return;
      } else if (key == "initial_frequency") {
        Config::setConfigValue("initial_frequency", getPart(cmd, 2));
      } else if (key == "request_interval") {
        Config::setConfigValue("request_interval", getPart(cmd, 2));
        transceiver.setRequestInterval(Config::REQUEST_INTERVAL);
        transceiver.restartRequest();
      } else if (key == "brightness") {
        Config::setConfigValue("brightness", getPart(cmd, 2));
        display.dim(Config::BRIGHTNESS);
      } else if (key == "restart_delay") {
        Config::setConfigValue("restart_delay", getPart(cmd, 2));
      } else if (key == "controller_addr") {
        Config::setConfigValue("controller_addr", getPart(cmd, 2));
      } else if (key == "civ_addr") {
        Config::setConfigValue("civ_addr", getPart(cmd, 2));
      } else if (key == "invert_koppler_output_c") {
        Config::setConfigValue("inv_coupler_out_c", getPart(cmd, 2));
      } else if (key == "invert_koppler_output_l") {
        Config::setConfigValue("inv_coupler_out_l", getPart(cmd, 2));
      } else if (key == "invert_koppler_output_io") {
        Config::setConfigValue("inv_coupler_out_io", getPart(cmd, 2));
      } else if (key == "scroll_width") {
        Config::setConfigValue("scroll_width", getPart(cmd, 2));
      } else if (key == "list_size") {
        Config::setConfigValue("list_size", getPart(cmd, 2));
      } else if (key == "step_bit_width_l") {
        Config::setConfigValue("step_bit_width_l", getPart(cmd, 2));
        resetRequired = true;
      } else if (key == "step_bit_width_c") {
        Config::setConfigValue("step_bit_width_c", getPart(cmd, 2));
        resetRequired = true;
      } else if (key == "step_c_min_value") {
        Config::setConfigValue("step_c_min_value", getPart(cmd, 2));
        resetRequired = true;
      } else if (key == "step_l_min_value") {
        Config::setConfigValue("step_l_min_value", getPart(cmd, 2));
        resetRequired = true;
      } else if (key == "pin_onewire") {
        Config::setConfigValue("pin_onewire", getPart(cmd, 2));
        resetRequired = true;
      } else if (key == "ic1") {
        Config::IC1[0] = Config::hex2Byte(getPart(cmd, 2));
        Config::IC1[1] = Config::hex2Byte(getPart(cmd, 3));
        Config::IC1[2] = Config::hex2Byte(getPart(cmd, 4));
        Config::IC1[3] = Config::hex2Byte(getPart(cmd, 5));
        Config::IC1[4] = Config::hex2Byte(getPart(cmd, 6));
        Config::IC1[5] = Config::hex2Byte(getPart(cmd, 7));
        Config::IC1[6] = Config::hex2Byte(getPart(cmd, 8));
        Config::IC1[7] = Config::hex2Byte(getPart(cmd, 9));
        resetRequired = true;
      } else if (key == "ic2") {
        Config::IC2[0] = Config::hex2Byte(getPart(cmd, 2));
        Config::IC2[1] = Config::hex2Byte(getPart(cmd, 3));
        Config::IC2[2] = Config::hex2Byte(getPart(cmd, 4));
        Config::IC2[3] = Config::hex2Byte(getPart(cmd, 5));
        Config::IC2[4] = Config::hex2Byte(getPart(cmd, 6));
        Config::IC2[5] = Config::hex2Byte(getPart(cmd, 7));
        Config::IC2[6] = Config::hex2Byte(getPart(cmd, 8));
        Config::IC2[7] = Config::hex2Byte(getPart(cmd, 9));
        resetRequired = true;
      } else if (key == "ic3") {
        Config::IC3[0] = Config::hex2Byte(getPart(cmd, 2));
        Config::IC3[1] = Config::hex2Byte(getPart(cmd, 3));
        Config::IC3[2] = Config::hex2Byte(getPart(cmd, 4));
        Config::IC3[3] = Config::hex2Byte(getPart(cmd, 5));
        Config::IC3[4] = Config::hex2Byte(getPart(cmd, 6));
        Config::IC3[5] = Config::hex2Byte(getPart(cmd, 7));
        Config::IC3[6] = Config::hex2Byte(getPart(cmd, 8));
        Config::IC3[7] = Config::hex2Byte(getPart(cmd, 9));
        resetRequired = true;
      }

    } else if (getPart(cmd, 0) == "append") {

      unsigned long freq = getPart(cmd, 1).toInt();  // frequency
      int c = getPart(cmd, 2).toInt();               // c
      int l = getPart(cmd, 3).toInt();               // l
      unsigned char io = getPart(cmd, 4).toInt();    // io

      Serial.println("save frequency:");
      Serial.println("freq: " + String(freq));
      Serial.println("c: " + String(c));
      Serial.println("l: " + String(l));
      Serial.println("io: " + String(io));

      lookuptable.save(freq, c, l, io);
      
    } else if (getPart(cmd, 0) == "dump") {

      String value = getPart(cmd, 1);
      Serial.println(value);
      lookuptable.dump(value);

    } else if (getPart(cmd, 0) == "rt") {
      transceiver.requestFrequency();
    } else if (getPart(cmd, 0) == "help") {
      showHelp();
    } else if (getPart(cmd, 0) == "update") {
      invalidate();
    }

    if (resetRequired) {
      Serial.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
      Serial.println("");
      Serial.println("Some changes require the system to be restarted.");
      Serial.println("Use 'reset' to restart the system!");
      Serial.println("");
      Serial.println("DON'T PANIC! SAVE YOUR FILE: 'config save'");
      Serial.println("!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!");
    }
  }
}

void setup() {

  // Initialize Serial port
  Serial.begin(Config::SERIAL_SPEED_SERIAL1);

  // Initialize Serial port for Transceiver
  Serial1.begin(Config::SERIAL_SPEED_SERIAL2);

  // Initialize serial poret for display
  Serial3.begin(Config::SERIAL_SPEED_SERIAL3);

  // wait for serial port
  while (!Serial) {}

  if (!SD.begin(Config::CHIP_SELECT_PIN)) {
    Serial.println("SD-Karte couldn't be initialized!");
    return;
  }

  Config::init(Config::FILENAME_CONFIG);
  Config::load();

  transceiver.ctrlAddress = Config::CONTROLLER_ADDR;
  transceiver.civAddress = Config::CIV_ADDR;

  // This call leads to a screen flicker
  // display.showPage(NAME_PAGE_SPLASH);
  display.writeDate(Config::DATE);
  display.writeVersion(Config::VERSION);

  OneWire net(Config::PIN_ONEWIRE);

  koppler.net = net;

  display.dim(Config::BRIGHTNESS);

  // initialize the system
  transceiver.setRequestInterval(Config::REQUEST_INTERVAL);
  transceiver.restartRequest();
  transceiver.requestFrequency();

  stepFrequency = Config::STEP_FREQUENCY_INITIAL_VALUE;
  step_lc = Config::STEP_LC_INITIAL_VALUE;
  transceiver.setFrequency(Config::INITIAL_FREQUENCY);

  display.writeStep(step_lc);
  display.writeStepFrequency(stepFrequency);

  // Needed to wait for answer of requestFrequency()
  delay(1000);
  transceiver.main();
  transceiver.main();
  lookuptable.setFrequency(transceiver.getFrequencyRounded());

  showHelp();

  if (lookuptable.getIO() == 1)
    display.showPage(Config::NAME_PAGE_AUTO_RIGHT);
  else
    display.showPage(Config::NAME_PAGE_AUTO_LEFT);
}

void loop() {

  console();

  if (display.pressed(Config::KEY_AUTO)) {
    Serial.println("Key pressed: KEY_AUTO");
    if (koppler.getIO() == 1)
      display.showPage(Config::NAME_PAGE_FREQ_RIGHT);
    else
      display.showPage(Config::NAME_PAGE_FREQ_LEFT);
  }
  if (display.pressed(Config::KEY_MAN1)) {
    Serial.println("Key pressed: KEY_MAN1");
    if (koppler.getIO() == 1)
      display.showPage(Config::NAME_PAGE_MAN_C_RIGHT);
    else
      display.showPage(Config::NAME_PAGE_MAN_C_LEFT);
  }
  if (display.pressed(Config::KEY_MAN2)) {
    Serial.println("Key pressed: KEY_MAN2");
    if (koppler.getIO() == 1) {
      display.showPage(Config::NAME_PAGE_AUTO_RIGHT);
    } else {
      display.showPage(Config::NAME_PAGE_AUTO_LEFT);
    }
    invalidate();
  }
  if (display.pressed(Config::KEY_SETUP)) {
    Serial.println("Key pressed: KEY_SETUP");
    display.showPage(Config::NAME_PAGE_SETUP);
  }
  if (display.pressed(Config::KEY_FREQUENCY_DOWN)) {
    Serial.println("Key pressed: KEY_FREQUENCY_DOWN");
    // manipulate rounded frequency
    unsigned long f = transceiver.getFrequency();
    f -= stepFrequency * 1000;
    transceiver.setFrequency(f);
  }
  if (display.pressed(Config::KEY_FREQUENCY_UP)) {
    Serial.println("Key pressed: KEY_FREQUENCY_UP");
    unsigned long f = transceiver.getFrequency();
    f += stepFrequency * 1000;
    transceiver.setFrequency(f);
  }
  if (display.pressed(Config::KEY_FREQUENCY_STEP)) {
    Serial.println("Key pressed: KEY_FREQUENCY_STEP");
    stepFrequency *= 10;
    if (stepFrequency > 1000)
      stepFrequency = 10;
    display.writeStepFrequency(stepFrequency);
  }
  if (display.pressed(Config::KEY_CAPACITANCE_UP)) {
    Serial.println("Key pressed: KEY_CAPACITANCE_UP");
    int c = koppler.getC() + step_lc;
    if (c + step_lc > pow(2, Config::STEP_BIT_WIDTH_C))
      c = 0;
    koppler.setC(c);
    display.writeC(c);
  }
  if (display.pressed(Config::KEY_CAPACITANCE_DOWN)) {
    Serial.println("Key pressed: KEY_CAPACITANCE_DOWN");
    int c = koppler.getC() - step_lc;
    if (c < 0)
      c = pow(2, Config::STEP_BIT_WIDTH_C);
    koppler.setC(c);
    display.writeC(c);
  }
  if (display.pressed(Config::KEY_INDUCTANCE_UP)) {
    Serial.println("Key pressed: KEY_INDUCTANCE_UP");
    int l = koppler.getL() + step_lc;
    if (l > pow(2, Config::STEP_BIT_WIDTH_L))
      l = 0;
    koppler.setL(l);
    display.writeL(l);
  }
  if (display.pressed(Config::KEY_INDUCTANCE_DOWN)) {
    Serial.println("Key pressed: KEY_INDUCTANCE_DOWN");
    int l = koppler.getL() - step_lc;
    if (l < 0)
      l = pow(2, Config::STEP_BIT_WIDTH_L);
    koppler.setL(l);
    display.writeL(l);
  }
  if (display.pressed(Config::KEY_STEP)) {
    Serial.println("Key pressed: KEY_STEP");
    step_lc *= 10;
    if (step_lc > 100)
      step_lc = 1;
    display.writeStep(step_lc);
  }
  if (display.pressed(Config::KEY_SWITCH_CAP_TRX)) {
    Serial.println("Key pressed: KEY_SWITCH_CAP_TRX");
    display.showPage(Config::NAME_PAGE_MAN_C_LEFT);
    koppler.setIO(2);
  }
  if (display.pressed(Config::KEY_SWITCH_CAP_ANT)) {
    Serial.println("Key pressed: KEY_SWITCH_CAP_ANT");
    display.showPage(Config::NAME_PAGE_MAN_C_RIGHT);
    koppler.setIO(1);
  }
  if (display.pressed(Config::KEY_STORE)) {
    display.startWaitPage();
    Serial.println("Key pressed: KEY_STORE");
    unsigned long frequency = transceiver.getFrequencyRounded();
    int c = koppler.getC();
    int l = koppler.getL();
    unsigned char io = koppler.getIO();
    lookuptable.save(frequency, c, l, io);
    display.endWaitPage();
  }
  if (display.pressed(Config::KEY_SETUP)) {
    display.showConfig();
  }
  if (display.pressed(Config::KEY_SEXIT)) {
    Serial.println("Key pressed: SEXIT");
  }
  if (display.pressed(Config::KEY_EXIT)) {
    Serial.println("Key pressed: EXIT");
    if (koppler.getIO() == 1) {
      display.showPage(Config::NAME_PAGE_AUTO_RIGHT);
    } else {
      display.showPage(Config::NAME_PAGE_AUTO_LEFT);
    }
    invalidate();
  }
  if (display.pressed(Config::KEY_RESET)) {
    Serial.println("Key pressed: RESET");
    reset();
  }
  if (display.pressed(Config::KEY_SPREV)) {
    display.configPrev();
  }
  if (display.pressed(Config::KEY_SNEXT)) {
    display.configNext();
  }
  if (display.pressed(Config::KEY_SPREV)) {
    Serial.println("Key pressed: SPREV");
  }
  if (display.pressed(Config::KEY_SNEXT)) {
    Serial.println("Key pressed: SNEXT");
  }
  if (transceiver.hasFrequencyChanged() || forceUpdate) {
    unsigned long fr = transceiver.getFrequencyRounded();
    unsigned long f = transceiver.getFrequency();

    // read parameter from lookuptable
    lookuptable.setFrequency(fr);
    int l = lookuptable.getL();
    int c = lookuptable.getC();
    unsigned char io = lookuptable.getIO();

    if (lookuptable.hasIOChanged() || forceUpdate) {
      if (transceiver.comesFrequencyFromTRX() || forceUpdate) {
        if (lookuptable.getIO() == 1)
          display.showPage(Config::NAME_PAGE_AUTO_RIGHT);
        else
          display.showPage(Config::NAME_PAGE_AUTO_LEFT);

      } else {
        if (lookuptable.getIO() == 1)
          display.showPage(Config::NAME_PAGE_FREQ_RIGHT);
        else
          display.showPage(Config::NAME_PAGE_FREQ_LEFT);
      }
      forceUpdate = false;
    }

    // set koppler
    koppler.setL(l);
    koppler.setC(c);
    koppler.setIO(io);

    display.writeFrequencyRounded(fr);
    display.writeFrequency(f);
    display.writeL(l);
    display.writeC(c);
  }

  display.isTransceiverConected(transceiver.isConnected());
  display.main();
  transceiver.main();
  koppler.main();
}
