#include <SD.h>
#include "Transceiver.h"
#include "Config.h"

Transceiver::Transceiver(int ctrlAddr, int civAddr) {
  
  this->ctrlAddress = ctrlAddr;
  this->civAddress = civAddr;
  this->frequency = 0;
  this->nextFrequency = 0;
  this->frequencyFromTRX = false;
  this->isRequestRequired = true;

}

bool Transceiver::comesFrequencyFromTRX(){
  return this->frequencyFromTRX;
}

bool Transceiver::isConnected(){
  return myRequestInterval>0;
}

void Transceiver::setFrequency(unsigned long value){
  nextFrequency = value;
  this->frequencyFromTRX = false;
}

unsigned long Transceiver::getFrequency(){
  return nextFrequency;
}

unsigned long Transceiver::getFrequencyRounded(){
  return ((nextFrequency + 5000) / 10000) * 10000;
}

bool Transceiver::hasFrequencyChanged(){
  return frequency != nextFrequency;
}

void Transceiver::main(){
  
  // Automatic request On/Off

  if( frequency != nextFrequency ){
    frequency = nextFrequency;
  }


  readTransceiver();

  if(!isRequestRequired){
    myRequestInterval = 0;
  }
  
  if(millis() - lastReceived > Config::RESTART_DELAY){
    myRequestInterval = requestInterval;
    isRequestRequired = true;
  }

  // interval
  unsigned long interval = millis();
  if(interval-lastInterval> myRequestInterval && myRequestInterval > 0){
    this->requestFrequency();
    lastInterval = interval;
  }

}

unsigned long Transceiver::getRequestInterval(){
  return myRequestInterval;
}

void Transceiver::restartRequest(){
  myRequestInterval = requestInterval;
  isRequestRequired = true;
}

void Transceiver::setRequestInterval(unsigned long interval){
  this->requestInterval = interval;
}

void Transceiver::readTransceiver(){

  bool foreignAnswer = false;

  if(!Serial1.available()){
    return;
  }

  unsigned char cmdEnd = 0xFD;
  unsigned int cmd[32];

  // read whole command from serial bus
  unsigned char qtyReceived = 0;
  int currByte;
  // Code smell: If the transceiver breaks during a command
  //             this ends in an endless loop!!
  while( true ){
    currByte = Serial1.read();
    if( currByte != -1){
      cmd[qtyReceived++] = currByte;
      if(currByte == cmdEnd)
        break;
    }
  }

  // Does the command begin with preamble
  if( cmd[0] != 0xFE && cmd[1] != 0xFE){
    return;
  }
  
  // Is the command for me?
  if( cmd[2] != ctrlAddress){
    //return;
    foreignAnswer = true;
  }
  
  //is the command from the correct sender
  if(cmd[3] != civAddress){
    return;
  }
  
  // Interpret the actual command
  switch(cmd[4]){
    case 0x00:  // Frequency is sent automatically from transceiver (transceive_on)
    case 0x03:  // Frequency was requested via CI-V
          
      if(qtyReceived != 11){
        return;
      }
      const unsigned long decMulti[] = { 1000000000, 100000000, 10000000, 1000000, 100000, 10000, 1000, 100, 10, 1 };  // Multiplikatorarray für Frequenzberechnung
      nextFrequency = 0;
      // FE FE 00 7A 03 xx xx xx xx xx FD (xx individual frequencydata)
      // FE FE 00 7A 03 <00 40 25 14 00> FD = 0014254000 Hz
      for (uint8_t i = 0; i < 5; i++) {                              // counting loop 0-5
        nextFrequency += (cmd[9 - i] >> 4) * decMulti[i * 2];        // read frequency from the buffer and convert to decimal
        nextFrequency += (cmd[9 - i] & 0x0F) * decMulti[i * 2 + 1];  // read frequency from the buffer and convert to decimal
      }
      this->frequencyFromTRX = true;

      break;
    default:
      Serial.print( "Command (" );
      Serial.print(cmd[4], HEX);
      Serial.println( ") not recognized!" );
      break;
  }

  if(foreignAnswer){
    this->isRequestRequired = false;
  }

  lastReceived = millis();
  
}

void Transceiver::requestFrequency(){
  uint8_t req[] = { 0xFE, 0xFE, civAddress, ctrlAddress, 0x03, 0xFD };
  for (uint8_t i = 0; i < sizeof(req); i++) {
    Serial1.write(req[i]);
    delay(5);  
  }
}